"""Class for finding the most likely url for an RSS/atom/other feed on a web page.

See https://github.com/dfm/feedfinder2 for other approaches to the same task.
"""
from urllib.parse import urljoin, urlparse, urlunparse

from bs4 import BeautifulSoup
import requests
from requests.adapters import HTTPAdapter
from requests.exceptions import InvalidSchema, RetryError
from requests.packages.urllib3.util.retry import Retry
from timeout_decorator import timeout, TimeoutError as CustomTimeoutError


def _is_feed_url(url):
    """Check if a url is a feed url with high confidence

    Wraps custom logic for high probability checks.

    Parameters
    ----------
    url : str
          Url that may or may not be a feed

    Returns
    -------
    boolean
        True if the string is a feed with high probability, or else False
    """
    endings = (
        '.rss',
        '.rdf',
        '.atom',
        '.xml',
    )
    url_lower = url.lower()
    return any(url_lower.endswith(ending) for ending in endings)


def _might_be_feed_url(url):
    """Check if a url might be a feed with moderate confidence

    A lower trust version of `_is_feed_url`
    Parameters
    ----------
    url : str
          Url that may or may not be a feed

    Returns
    -------
    boolean
        True if the string is a feed with reasonable probability, or else False
    """
    substrings = (
        'rss',
        'rdf',
        'atom',
        'xml',
        'feed'
    )
    url_lower = url.lower()
    return any(substring in url_lower for substring in substrings)


def default_fetch_function(url):
    """Default function to fetch text from url

    There are some strong choices on how to handle errors in `FeedSeeker`. Use this function
    as an example of how to make a new fetch function. Note that the function may be used
    to attempt to fetch urls that do not exist, so be thoughtful about what exceptions to throw!

    Parameters
    ----------
    url : string
        A url for a webpage

    Returns
    ------
    str
        Text of the html from the url
    """

    session = requests.Session()

    retries = Retry(total=5, backoff_factor=0.1, status_forcelist=[500, 502, 503, 504])
    session.mount(url, HTTPAdapter(max_retries=retries))
    try:
        response = session.get(url)
        if response.ok:
            return response.text
        else:
            return ''

    # ConnectionError for 404s, InvalidSchema for email addresses
    except (requests.ConnectionError, InvalidSchema, RetryError):
        return ''


class FeedSeeker(object):
    """A class to find possible RSS/Atom feeds on a web page.

    Usually easier to use `find_feed_url` or `generate_feed_urls`.
    The class defines a few methods for discovering links on a page, from
    first looking for standard feed links, then looking for *any* link, and then
    guessing at a few urls that are commonly used for feeds.  See
    `find_link_feeds`, `find_anchor_feeds`, and `guess_feed_links`, respectively.

    The class is used by either asking for a single feed, or all feeds.  All web
    fetching is deferred until needed, so it is typically much faster to only
    get a single feed, or to stop iterating over all feeds once a condition is
    satisfied.
    """
    def __init__(self, url, html=None, fetcher=None):
        """Initialization

        Parameters
        ----------
        url : str
              A url that resolves to the webpage in question

        html : str (optional)
              To save a second web fetch, the raw html can be supplied
        fetcher : function (optional)
              A function that accepts a url and returns text. See `default_fetch_function`
              for how to write a custom fetcher.
        """
        self.url = url
        self._html = html
        self._soup = None
        self.fetcher = fetcher or default_fetch_function

    @property
    def html(self):
        """String of the html of the underlying site."""
        if self._html is None:
            self._html = self.fetcher(self.url)
        return self._html

    @property
    def soup(self):
        """BeautifulSoup representation of the data."""
        if self._soup is None:
            self._soup = BeautifulSoup(self.html, 'lxml')
        return self._soup

    def clean_url(self):
        """Remove query arguments from a url."""
        parsed = urlparse(self.url)
        return urlunparse(parsed._replace(query=''))

    def _should_continue(self, seen, max_links):
        """Helper to short-circuit spidering
        Parameters
        ----------
        seen : set
            List of urls that have already been checked
        max_links : int or None
            Maximum number of links to check

        Returns
        ------
        boolean
            True if no stop condition has been met, False otherwise
        """
        if max_links is not None and len(seen) >= max_links:
            return False
        return True

    def generate_feed_urls(self, spider=0, max_links=None):
        """Generates an iterator of possible feeds, in rough order of likelihood.

        Parameters
        ----------
        spider : int (optional)
              How many times to restart the seeker on links with the same hostname on this page
        max_links : int (optional)
              Maximum links to check as feeds, to limit spidering complexity. Defaults to `None`,
              for unlimited.

        Yields
        ------
            urls of possible feeds
        """
        for url, _ in self._generate_feed_urls(spider=spider, max_links=max_links):
            yield url

    def _generate_feed_urls(self, spider=0, seen=None, max_links=None):
        """Internal function that actually does the work for `generate_feed_urls`

        There are some recursive calls keeping track of already seen urls, and it was easier
        to do it this way.

        Parameters
        ----------
        spider : int (optional)
              How many times to restart the seeker on links with the same hostname on this page
        seen : set
            (Optional) list of urls to not produce. May be used as a blacklist.
        max_links : int (optional)
              Maximum links to check as feeds, to limit spidering complexity. Defaults to `None`,
              for unlimited.

        Yields
        ------
            (string, set)
            urls of possible feeds, and all the urls already seen
        """
        if seen is None:
            seen = set()

        if not self.html:
            return

        if self.is_feed() and self.url not in seen:
            seen.add(self.url)
            yield self.url, seen
            return

        for url_fn in (self.find_link_feeds, self.find_anchor_feeds, self.guess_feed_links):
            for url in url_fn():
                if url not in seen:
                    seen.add(url)
                    if not self._should_continue(seen, max_links):
                        return
                    if FeedSeeker(url).is_feed():
                        yield url, seen

        if spider > 0:
            for internal_link in self.find_internal_links():
                spider_seeker = FeedSeeker(internal_link, html=None, fetcher=self.fetcher)
                kwargs = {
                    'spider': spider - 1,
                    'seen': seen,
                    'max_links': max_links,
                }
                for url, seen in spider_seeker._generate_feed_urls(**kwargs):
                    yield url, seen

    def find_feed_url(self, spider=0, max_links=None):
        """Fine the single most likely url as a feed for the page, or None.

        Parameters
        ----------
        spider : int (optional)
              How many times to restart the seeker on links with the same hostname on this page
        max_links : int (optional)
              Maximum links to check as feeds, to limit spidering complexity. Defaults to `None`,
              for unlimited.

        Returns
        -------
        str
            The most likely url to have a feed for this page
        """

        try:
            return next(self.generate_feed_urls(spider=spider, max_links=max_links))
        except StopIteration:
            return None

    def is_feed(self):
        """Check if the site is a feed.

        Logic is to make sure there is no <html> tag, and there is some <rss> tag or similar.
        """
        invalid_tags = ('head',)
        if any(self.soup.find(tag) for tag in invalid_tags):
            return False

        valid_tags = ('rss', 'rdf', 'feed',)
        return any(self.soup.find(tag) for tag in valid_tags)

    def find_link_feeds(self):
        """Uses <link> tags to extract feeds

        for example:
            <link type="application/rss+xml" href="/might/be/relative.rss"></link>
        """
        valid_types = [
            "application/rss+xml",
            "text/xml",
            "application/atom+xml",
            "application/x.atom+xml",
            "application/x-atom+xml"
        ]

        for link in self.soup.find_all('link', type=valid_types):
            url = link.get('href')
            if url:
                yield urljoin(base=self.clean_url(), url=url)

    def find_internal_links(self):
        """Finds <a></a> tags to internal pages on the same domain that may have a feed.

        For example, this may find the homepage, or an index page.
        """
        parsed_url = urlparse(self.clean_url())
        parts = set(filter(None, parsed_url.path.split('/')))
        possible_links = []
        for link_node in self.soup.find_all('a', href=True):
            link = link_node.get('href')
            parsed_link = urlparse(link)
            if not parsed_link.hostname:
                parsed_link = parsed_link._replace(netloc=parsed_url.hostname,
                                                   scheme=parsed_url.scheme)
                link = urlunparse(parsed_link)
            if parsed_link.hostname == parsed_url.hostname:
                might_be_feed = any(check(link) for check in (_is_feed_url, _might_be_feed_url))
                link_parts = set(filter(None, parsed_link.path.split('/')))
                similarity = len(parts.intersection(link_parts)) + might_be_feed * len(parts)
                possible_links.append((link, similarity))
        return [link for link, _ in sorted(set(possible_links), key=lambda j: (-j[1], len(j[0])))]


    def find_anchor_feeds(self):
        """Uses <a></a> tags to extract feeds

        for example
            <a href="https://www.whatever.com/rss"></a>
        """
        # This is outer loop so that most likely links
        # are produced first
        for url_filter in (_is_feed_url, _might_be_feed_url):
            for link in self.soup.find_all('a', href=True):
                url = link.get('href')
                if url_filter(url):
                    yield urljoin(base=self.clean_url(), url=url)

    def guess_feed_links(self):
        """Iterates common locations to find feeds.  These urls probably do not exist, but might

        Manual overrides should be added here.  For example, if foo.com has their rss feed at
        foo.com/here/for/reasons.rss, add 'here/for/reasons.rss' to the suffixes.
        """
        suffixes = (
            # Generic suffixes
            'index.xml', 'atom.xml', 'feeds', 'feeds/default', 'feed', 'feed/default',
            'feeds/posts/default/', '?feed=rss', '?feed=atom', '?feed=rss2', '?feed=rdf', 'rss',
            'atom', 'rdf', 'index.rss', 'index.rdf', 'index.atom',
            '?type=100',  # Typo3 RSS URL
            '?format=feed&type=rss',  # Joomla RSS URL
            'feeds/posts/default',  # Blogger.com RSS URL
            'data/rss',  # LiveJournal RSS URL
            'rss.xml',  # Posterous.com RSS feed
            'articles.rss', 'articles.atom',  # Patch.com RSS feeds
        )
        for suffix in suffixes:
            yield urljoin(base=self.clean_url(), url=suffix)


def find_feed_url(url, html=None, spider=0, max_time=None, max_links=None):
    """Find the single most likely feed url for a page.

    Parameters
    ----------
    url : str
          A url that resolves to the webpage in question
    html : str (optional)
          To save a second web fetch, the raw html can be supplied
    spider : int (optional)
          How many times to restart the seeker on links with the same hostname on this page
    max_time : float (optional)
          Give up after a certain amount of time. This is a lower limit, in that time
          is checked *after* each request returns. Defaults to `None` for unlimited. Will
          throw a TimeoutError if reached
    max_links : int (optional)
          Maximum links to check as feeds, to limit spidering complexity. Defaults to `None`,
          for unlimited.


    Returns
    -------
    str or None
       A url pointing to the most likely feed, if it exists.
    """
    time_out_fn = timeout(seconds=max_time)
    decorated_fn = time_out_fn(FeedSeeker(url, html).find_feed_url)
    try:
        return decorated_fn(spider=spider, max_links=max_links)
    except CustomTimeoutError as exc:
        raise TimeoutError('Max time ({}s) reached'.format(str(max_time))) from exc

def generate_feed_urls(url, html=None, spider=0, max_time=None, max_links=None):
    """Find all feed urls for a page.

    Parameters
    ----------
    url : str
          A url that resolves to the webpage in question
    html : str (optional)
          To save a second web fetch, the raw html can be supplied
    spider : int (optional)
          How many times to restart the seeker on links with the same hostname on this page
    max_time : float (optional)
          Give up after a certain amount of time. This is a lower limit, in that time
          is checked *after* each request returns. Defaults to `None` for unlimited. Throws
          a TimeoutError when the time is reached
    max_links : int (optional)
          Maximum links to check as feeds, to limit spidering complexity. Defaults to `None`,
          for unlimited.

    Yields
    ------
    str or None
       A url pointing to a feed associated with the page
    """
    @timeout(seconds=max_time, use_signals=False, timeout_exception=TimeoutError)
    def _generate_feed_urls(url, html=None, spider=0, max_links=None):
        """Helper for generate_feed_urls.

        This is needed because the timeout decorator does not work well with generators
        """
        return FeedSeeker(url, html).generate_feed_urls(spider=spider, max_links=max_links)

    return _generate_feed_urls(url, html=html, spider=spider, max_links=max_links)




import time

import pytest
import responses

from feed_seeker import feed_seeker


def test__is_feed_url():
    assert not feed_seeker._is_feed_url('nytimes.com')
    assert feed_seeker._is_feed_url('nytimes.rss')
    assert not feed_seeker._is_feed_url('rssnews.com')


def test__might_be_feed_url():
    assert not feed_seeker._might_be_feed_url('nytimes.com')
    assert feed_seeker._might_be_feed_url('nytimes.rss')
    assert feed_seeker._might_be_feed_url('rssnews.com')


@responses.activate
def test_find_feed_url_max_time():
    max_time = 0.5

    def request_callback(request):
        time.sleep(2 * max_time)
        return (200, {}, '')

    url = 'http://nopenopenope.nope'
    responses.add_callback(responses.GET, url, callback=request_callback)

    with pytest.raises(TimeoutError):
        feed_seeker.find_feed_url(url, max_time=max_time)

    def request_callback(request):
        return (200, {}, '')

    # make sure it doesn't always raise
    url = 'http://yupyupyup.yup'
    responses.add_callback(responses.GET, url, callback=request_callback)
    feed_seeker.find_feed_url(url, max_time=max_time)


class TestFeedSeeker(object):
    def setup_method(self):
        responses.start()

        self.base_url = 'http://nopenopenope.nope'
        self.regular_html_template = "<html><head>{head}</head><body>{body}</body></html>"
        self.rss_feed_template = '<link type="application/rss+xml" href="{}" />'
        self.regular_feed_page = '<?xml version="1.0"?> <rss version="2.0"></rss>'

    def teardown_method(self):
        responses.stop()
        responses.reset()

    def test_is_feed(self):
        finder = feed_seeker.FeedSeeker(self.base_url, html=self.regular_feed_page)
        assert finder.is_feed()
        finder = feed_seeker.FeedSeeker(self.base_url, html=self.regular_html_template)
        assert not finder.is_feed()

    def test_html_property(self):
        responses.add(responses.GET, self.base_url, body=self.regular_html_template, status=200)
        finder = feed_seeker.FeedSeeker(self.base_url)
        found_html = finder.html
        assert found_html == self.regular_html_template

    def generate_responses(self):
        feeds = (
            '/get_your_news_here.html',  # will be found in <head>
            '{}/atom.rss'.format(self.base_url),  # will be found in <body>
            '/index.xml',  # hidden!  will still get spotted
        )

        non_feeds = (
            '/not_a_feed.rss',  # will be in <head>
            '/rss.atom',  # will be in <body>
            '/atom.xml',  # hidden, but not 404
        )

        html = self.regular_html_template.format(
            head=self.rss_feed_template.format(feeds[0]),
            body='<a href="{}"></a>'.format(feeds[1])
        )

        assert feeds[0] in html
        assert feeds[1] in html
        assert feeds[2] not in html

        responses.add(responses.GET, self.base_url, body=html, status=200)

        for feed in feeds:
            if not feed.startswith(self.base_url):
                feed = self.base_url + feed
            responses.add(responses.GET, feed, body=self.regular_feed_page, status=200)

        for feed in non_feeds[:-1]:
            responses.add(responses.GET, self.base_url + feed,
                          body=self.regular_html_template, status=200)

        # handle bad statuses
        responses.add(responses.GET, self.base_url + non_feeds[-1],
                      body=self.regular_html_template, status=500)

        return feeds, non_feeds

    def test_spider(self):
        steps = 5
        urls, feeds = [], []
        for step in range(steps):
            html = self.regular_html_template.format(
                head=self.rss_feed_template.format('/{}.rss'.format(step)),
                body='<a href="/{}.html"'.format(step + 1))
            urls.append(self.base_url + '/{}.html'.format(step))
            feeds.append(self.base_url + '/{}.rss'.format(step))
            responses.add(responses.GET, urls[-1], body=html, status=200)
            responses.add(responses.GET, feeds[-1], body=self.regular_feed_page, status=200)

        for j in range(steps - 1):
            found_feeds = list(feed_seeker.generate_feed_urls(urls[0], spider=j))
            assert feeds[:j + 1] == found_feeds

    def test_generate_feed_urls(self):
        feeds, _ = self.generate_responses()

        finder = feed_seeker.FeedSeeker(self.base_url)
        found_feeds = list(finder.generate_feed_urls())

        assert len(found_feeds) == len(feeds)

    def test_generate_feed_urls_max_links(self):
        feeds, _ = self.generate_responses()

        finder = feed_seeker.FeedSeeker(self.base_url)
        max_links = 2
        found_feeds = list(finder.generate_feed_urls(max_links=max_links))

        assert len(found_feeds) > 0
        assert len(found_feeds) <= max_links < len(feeds)

    def test_generate_feed_urls_function(self):
        feeds, _ = self.generate_responses()

        found_feeds = list(feed_seeker.generate_feed_urls(self.base_url))

        assert len(found_feeds) == len(feeds)

    def test_generate_feed_urls_not_a_page(self):
        feeds, _ = self.generate_responses()

        finder = feed_seeker.FeedSeeker(self.base_url + '/what_is_this_even')
        found_feeds = list(finder.generate_feed_urls())

        assert len(found_feeds) == 0

    def test_generate_feed_urls_on_feed(self):
        feeds, _ = self.generate_responses()

        finder = feed_seeker.FeedSeeker(self.base_url + feeds[0])
        found_feeds = list(finder.generate_feed_urls())

        assert len(found_feeds) == 1

    def test_find_feed_url(self):
        feeds, _ = self.generate_responses()

        url = feed_seeker.find_feed_url(self.base_url)
        assert url == self.base_url + feeds[0]

    def test_find_link_feeds(self):
        num_feeds = 4
        feed_urls = []
        # note that we do NOT eliminate duplicates at this level
        for _ in range(num_feeds):
            feed_urls.append(self.rss_feed_template.format('http://whatever.com'))

        html = self.regular_html_template.format(head='\n'.join(feed_urls), body='')
        finder = feed_seeker.FeedSeeker(self.base_url, html=html)
        assert len(list(finder.find_link_feeds())) == num_feeds
        assert len(list(finder.find_anchor_feeds())) == 0

    def test_find_anchor_feeds(self):
        num_feeds = 4
        feed_urls = []
        # we should find these four links
        for feed in range(num_feeds):
            feed_urls.append('<a href="http://{}.rss"></a>'.format(feed))

        # but will not flag this one, since it does not look like a feed
        feed_urls.append('<a href="https://not_an_example.com"></a>')

        html = self.regular_html_template.format(head='', body='\n'.join(feed_urls))
        finder = feed_seeker.FeedSeeker(self.base_url, html=html)
        assert len(list(finder.find_link_feeds())) == 0
        assert len(set(finder.find_anchor_feeds())) == num_feeds

    def test_find_internal_links(self):
        _ = self.generate_responses()
        finder = feed_seeker.FeedSeeker(self.base_url, html=None)
        internal_links = finder.find_internal_links()
        assert len(internal_links) == 1 # from `self.generate_responses`

    def test_guess_feed_links(self):
        # even empty page has some guesses
        finder = feed_seeker.FeedSeeker(self.base_url, html=self.regular_html_template)
        guessed_links = list(finder.guess_feed_links())
        assert len(guessed_links) > 0
        for feed_link in guessed_links:
            assert self.base_url in feed_link

    def test_empty_page(self):
        finder = feed_seeker.FeedSeeker(self.base_url, html=self.regular_html_template)
        # Page has no links, so should fail
        assert finder.find_feed_url() is None



from codecs import open
from os.path import join, abspath, dirname
from setuptools import setup, find_packages

here = abspath(dirname(__file__))

# Get the long description from the README file
with open(join(here, 'README.rst'), encoding='utf-8') as buff:
    long_description = buff.read()

requirements_file = join(here, 'requirements.txt')

with open(requirements_file) as f:
    install_reqs = f.read().splitlines()

setup(
    name='feed_seeker',
    version='0.0.1',
    description='Extract rss, atom, and other feeds from webpages',
    long_description=long_description,
    author='Colin Carroll',
    author_email='ccarroll@mit.edu',
    url='https://github.com/mitmedialab/feed_seeker',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
    ],
    packages=find_packages(exclude=['test']),
    install_requires=install_reqs,
    include_package_data=True,
)
